@@ -1,5 +1,6 @@
 {-# LANGUAGE CPP                #-}
 {-# LANGUAGE DeriveAnyClass     #-}
+{-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE DeriveGeneric      #-}
 {-# LANGUAGE DeriveLift         #-}
 {-# LANGUAGE DerivingStrategies #-}
@@ -50,6 +51,10 @@ module Data.Versions
   , Chunk(..)
   , MChunk(..)
   , VSep(..)
+    -- ** Compile-time Constructors
+  , versioningQ, semverQ, versionQ, messQ, pvpQ
+    -- ** Conversions
+  , semverToVersion, versionToMess, versionToPvp
     -- * Parsing Versions
   , ParsingError
   , versioning, semver, pvp, version, mess
@@ -76,6 +81,7 @@ import qualified Control.Applicative.Combinators.NonEmpty as PC
 import           Control.DeepSeq
 import           Control.Monad (unless, void)
 import           Data.Char (isAlpha, isAlphaNum)
+import           Data.Data (Data)
 import           Data.Foldable (fold)
 import           Data.Hashable (Hashable)
 import           Data.List (intersperse)
@@ -85,12 +91,13 @@ import           Data.Maybe (fromMaybe, listToMaybe, mapMaybe)
 import           Data.Text (Text)
 import qualified Data.Text as T
 import           Data.Void (Void)
+import           Data.Word (Word64)
 import           GHC.Generics (Generic)
-import           Language.Haskell.TH.Syntax (Lift)
+import           Language.Haskell.TH (Exp, Q)
+import           Language.Haskell.TH.Syntax (Lift(..))
 import           Text.Megaparsec hiding (chunk)
 import           Text.Megaparsec.Char
 import qualified Text.Megaparsec.Char.Lexer as L
-import           Text.Megaparsec.Char.Lexer (decimal)
 
 ---
 
@@ -98,8 +105,8 @@ import           Text.Megaparsec.Char.Lexer (decimal)
 -- This allows each subtype to have its own parser, and for said parsers to be
 -- composed. This is useful for specifying custom behaviour for when a certain
 -- parser fails.
-data Versioning = Ideal SemVer | General Version | Complex Mess
-  deriving (Eq, Show, Generic, NFData, Hashable, Lift)
+data Versioning = Ideal !SemVer | General !Version | Complex !Mess
+  deriving (Eq, Show, Generic, NFData, Hashable, Lift, Data)
 
 -- | Short-hand for detecting a `SemVer`.
 isIdeal :: Versioning -> Bool
@@ -126,19 +133,19 @@ isComplex _           = False
 -- and @Complex@ is well defined for the same reason. This implies comparison of
 -- @Ideal@ and @Complex@ is also well-defined.
 instance Ord Versioning where
-  compare (Ideal s)     (Ideal s')   = compare s s'
-  compare (General v)   (General v') = compare v v'
-  compare (Complex m)   (Complex m') = compare m m'
-  compare (Ideal s)     (General v)  = semverAndVer s v
-  compare (General v)   (Ideal s)    = opposite $ semverAndVer s v
-  compare (General v)   (Complex m)  = compare (mFromV v) m
-  compare (Complex m)   (General v)  = opposite $ compare (mFromV v) m
-  compare (Ideal s)     (Complex m)  = semverAndMess s m
-  compare (Complex m)   (Ideal s)    = opposite $ semverAndMess s m
+  compare (Ideal s)   (Ideal s')   = compare s s'
+  compare (General v) (General v') = compare v v'
+  compare (Complex m) (Complex m') = compare m m'
+  compare (Ideal s)   (General v)  = semverAndVer s v
+  compare (General v) (Ideal s)    = opposite $ semverAndVer s v
+  compare (General v) (Complex m)  = compare (versionToMess v) m
+  compare (Complex m) (General v)  = opposite $ compare (versionToMess v) m
+  compare (Ideal s)   (Complex m)  = semverAndMess s m
+  compare (Complex m) (Ideal s)    = opposite $ semverAndMess s m
 
 -- | Convert a `SemVer` to a `Version`.
-vFromS :: SemVer -> Version
-vFromS (SemVer ma mi pa re me) =
+semverToVersion :: SemVer -> Version
+semverToVersion (SemVer ma mi pa re me) =
   Version
   { _vEpoch = Nothing
   , _vChunks = Chunks $ Numeric ma :| [Numeric mi, Numeric pa]
@@ -146,8 +153,8 @@ vFromS (SemVer ma mi pa re me) =
   , _vRel = re }
 
 -- | Convert a `Version` to a `Mess`.
-mFromV :: Version -> Mess
-mFromV (Version me (Chunks v) r _) = case me of
+versionToMess :: Version -> Mess
+versionToMess (Version me (Chunks v) r _) = case me of
   Nothing -> f
   Just e  ->
     let cs = (:| []) . MDigit e $ showt e
@@ -163,6 +170,17 @@ mFromV (Version me (Chunks v) r _) = case me of
       where
         ms = NEL.map toMChunk cs
 
+-- | Convert a `Version` to a `PVP`. Fails if there is an epoch present, but
+-- otherwise ignores the `Release` and other metadata. Naturally it also fails
+-- if any of the version components contain any non-digits.
+versionToPvp :: Version -> Maybe PVP
+versionToPvp (Version (Just _) _ _ _) = Nothing
+versionToPvp (Version Nothing (Chunks cs) _ _) = PVP <$> traverse f cs
+  where
+    f :: Chunk -> Maybe Word
+    f (Numeric w)  = Just w
+    f (Alphanum _) = Nothing
+
 semverAndVer :: SemVer -> Version -> Ordering
 -- A `Version` with a non-zero epoch value is automatically greater than any
 -- `SemVer`.
@@ -231,7 +249,7 @@ semverAndMess s@(SemVer ma mi pa _ _) m = case compare ma <$> messMajor m of
           EQ -> GT
   where
     fallback :: Ordering
-    fallback = compare (General $ vFromS s) (Complex m)
+    fallback = compare (General $ semverToVersion s) (Complex m)
 
 instance Semantic Versioning where
   major f (Ideal v)   = Ideal   <$> major f v
@@ -386,7 +404,7 @@ data SemVer = SemVer
   , _svPatch  :: !Word
   , _svPreRel :: !(Maybe Release)
   , _svMeta   :: !(Maybe Text) }
-  deriving stock (Show, Generic, Lift)
+  deriving stock (Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 -- | Two SemVers are equal if all fields except metadata are equal.
@@ -427,7 +445,7 @@ instance Semantic SemVer where
 
 -- | `Chunk`s have comparison behaviour according to SemVer's rules for preleases.
 newtype Release = Release (NonEmpty Chunk)
-  deriving stock (Eq, Show, Read, Generic, Lift)
+  deriving stock (Eq, Show, Read, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 instance Ord Release where
@@ -458,8 +476,8 @@ instance Ord Release where
 -- r3
 -- 0rc1-abc3
 -- @
-data Chunk = Numeric Word | Alphanum Text
-  deriving stock (Eq, Show, Read, Generic, Lift)
+data Chunk = Numeric !Word | Alphanum !Text
+  deriving stock (Eq, Show, Read, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 toMChunk :: Chunk -> MChunk
@@ -525,7 +543,7 @@ singleDigitLenient (Alphanum s) = hush $ parse unsignedP "Single Digit Lenient"
 --    change. The spec otherwise designates no special meaning to components
 --    past the MINOR position.
 newtype PVP = PVP { _pComponents :: NonEmpty Word }
-  deriving stock (Eq, Ord, Show, Generic)
+  deriving stock (Eq, Ord, Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 instance Semantic PVP where
@@ -577,7 +595,7 @@ data Version = Version
   , _vChunks :: !Chunks
   , _vRel    :: !(Maybe Release)
   , _vMeta   :: !(Maybe Text) }
-  deriving stock (Eq, Show, Generic, Lift)
+  deriving stock (Eq, Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 -- | Customized. As in SemVer, metadata is ignored for the purpose of
@@ -620,7 +638,7 @@ instance Semantic Version where
   {-# INLINE meta #-}
 
   semantic f (Version _ (Chunks (Numeric a :| Numeric b : Numeric c : _)) rs me) =
-    vFromS <$> f (SemVer a b c rs me)
+    semverToVersion <$> f (SemVer a b c rs me)
   semantic _ v = pure v
   {-# INLINE semantic #-}
 
@@ -631,7 +649,7 @@ epoch f v = fmap (\ve -> v { _vEpoch = ve }) (f $ _vEpoch v)
 
 -- | `Chunk`s that have a comparison behaviour specific to `Version`.
 newtype Chunks = Chunks (NonEmpty Chunk)
-  deriving stock (Eq, Show, Generic, Lift)
+  deriving stock (Eq, Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 instance Ord Chunks where
@@ -653,13 +671,13 @@ instance Ord Chunks where
 -- it could be, alongside the original text it came from. This preserves both
 -- `Ord` and pretty-print behaviour for versions like @1.003.0@.
 data MChunk
-  = MDigit Word Text
+  = MDigit !Word !Text
   -- ^ A nice numeric value.
-  | MRev Word Text
+  | MRev !Word !Text
   -- ^ A numeric value preceeded by an @r@, indicating a revision.
-  | MPlain Text
+  | MPlain !Text
   -- ^ Anything else.
-  deriving stock (Eq, Show, Generic, Lift)
+  deriving stock (Eq, Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 instance Ord MChunk where
@@ -688,7 +706,7 @@ mchunkText (MPlain t)   = t
 -- internal tests show consistency. `messMajor`, etc., are used internally where
 -- appropriate to enhance accuracy.
 data Mess = Mess !(NonEmpty MChunk) !(Maybe (VSep, Mess))
-  deriving stock (Eq, Show, Generic, Lift)
+  deriving stock (Eq, Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
 -- | Try to extract the "major" version number from `Mess`, as if it were a
@@ -750,7 +768,7 @@ instance Semantic Mess where
 
   -- | Good luck.
   semantic f (Mess (MDigit t0 _ :| MDigit t1 _ : MDigit t2 _ : _) _) =
-    mFromV . vFromS <$> f (SemVer t0 t1 t2 Nothing Nothing)
+    versionToMess . semverToVersion <$> f (SemVer t0 t1 t2 Nothing Nothing)
   semantic _ v = pure v
   {-# INLINE semantic #-}
 
@@ -763,9 +781,44 @@ instance Semantic Mess where
 -- * A plus (+). Stop using this outside of metadata if you are. Example: @10.2+0.93+1-1@
 -- * An underscore (_). Stop using this if you are.
 data VSep = VColon | VHyphen | VPlus | VUnder | VTilde
-  deriving stock (Eq, Show, Generic, Lift)
+  deriving stock (Eq, Show, Generic, Lift, Data)
   deriving anyclass (NFData, Hashable)
 
+-- | Parse a `Versioning` at compile time.
+versioningQ :: Text -> Q Exp
+versioningQ nm =
+  case versioning nm of
+    Left err -> fail (errorBundlePretty err)
+    Right v  -> lift v
+
+-- | Parse a `SemVer` at compile time.
+semverQ :: T.Text -> Q Exp
+semverQ nm =
+  case semver nm of
+    Left err -> fail (errorBundlePretty err)
+    Right v  -> lift v
+
+-- | Parse a `Version` at compile time.
+versionQ :: T.Text -> Q Exp
+versionQ nm =
+  case version nm of
+    Left err -> fail (errorBundlePretty err)
+    Right v  -> lift v
+
+-- | Parse a `Mess` at compile time.
+messQ :: T.Text -> Q Exp
+messQ nm =
+  case mess nm of
+    Left err -> fail (errorBundlePretty err)
+    Right v  -> lift v
+
+-- | Parse a `PVP` at compile time.
+pvpQ :: T.Text -> Q Exp
+pvpQ nm =
+  case pvp nm of
+    Left err -> fail (errorBundlePretty err)
+    Right v  -> lift v
+
 --------------------------------------------------------------------------------
 -- Parsing
 
@@ -797,7 +850,17 @@ semver'' = SemVer <$> majorP <*> minorP <*> patchP <*> optional releaseP <*> opt
 
 -- | Parse a group of digits, which can't be lead by a 0, unless it is 0.
 unsignedP :: Parsec Void Text Word
-unsignedP = (0 <$ char '0') <|> decimal
+unsignedP = asWord64 >>= convertOrFail
+  where
+    asWord64 :: Parsec Void Text Word64
+    asWord64 = (0 <$ char '0') <|> L.decimal
+
+    convertOrFail :: Word64 -> Parsec Void Text Word
+    convertOrFail w | w > bound = fail $ "Value (" ++ show w ++ ") larger than Word size: " ++ show bound
+                    | otherwise = pure $ fromIntegral w
+      where
+        bound :: Word64
+        bound = fromIntegral (maxBound :: Word)
 
 majorP :: Parsec Void Text Word
 majorP = unsignedP <* char '.'
@@ -851,7 +914,7 @@ pvp = parse (pvp' <* eof) "PVP"
 
 -- | Internal megaparsec parser of `pvp`.
 pvp' :: Parsec Void Text PVP
-pvp' = L.lexeme space (PVP . NEL.fromList <$> L.decimal `sepBy` char '.')
+pvp' = L.lexeme space (PVP <$> unsignedP `PC.sepBy1` char '.')
 
 -- | Parse a (General) `Version`, as defined above.
 version :: Text -> Either ParsingError Version
@@ -865,7 +928,7 @@ version'' :: Parsec Void Text Version
 version'' = Version <$> optional (try epochP) <*> chunksP <*> optional releaseP <*> optional metaData
 
 epochP :: Parsec Void Text Word
-epochP = read <$> (some digitChar <* char ':')
+epochP = unsignedP <* char ':'
 
 chunksP :: Parsec Void Text Chunks
 chunksP = Chunks <$> chunkWithoutHyphensP `PC.sepBy1` char '.'
@@ -969,7 +1032,7 @@ hush :: Either a b -> Maybe b
 hush (Left _)  = Nothing
 hush (Right b) = Just b
 
-data These a b = This a | That b | Both a b
+data These a b = This !a | That !b | Both !a !b
 
 zipLongest :: [a] -> [b] -> [These a b]
 zipLongest [] []         = []
