@@ -18,7 +18,6 @@ import           Test.Tasty.HUnit
 import           Text.Megaparsec
 import           Text.Megaparsec.Char
 import           Text.Printf (printf)
-import           TH (thVer)
 
 ---
 
@@ -27,11 +26,15 @@ goodVers :: [T.Text]
 goodVers = [ "1", "1.2", "1.0rc0", "1.0rc1", "1.1rc1", "1.58.0-3",  "44.0.2403.157-1"
            , "0.25-2",  "8.u51-1", "21-2", "7.1p1-1", "20150826-1", "1:0.10.16-3"
            , "1.11.0.git.20200404-1", "1.11.0+20200830-1", "1:3.20", "9.2.1.b-debug+lol"
+           , "0:1.2.3"
            ]
 
 badVers :: [T.Text]
 badVers = ["", "1.2 "]
 
+badPVP :: [T.Text]
+badPVP = ["", "clc237", "abc"]
+
 messes :: [T.Text]
 messes = [ "10.2+0.93+1-1", "003.03-3", "002.000-7", "20.26.1_0-2", "1.6.0a+2014+m872b87e73dfb-1"
          , "1.3.00.16851-1", "5.2.458699.0906-1", "12.0.0-3ubuntu1~20.04.5" ]
@@ -42,7 +45,12 @@ messComps = [ "10.2+0.93+1-1", "10.2+0.93+1-2", "10.2+0.93+2-1"
             ]
 
 badSemVs :: [T.Text]
-badSemVs = [ "1", "1.2", "a.b.c", "1.01.1", "1.2.3+a1b!2c3.1", "", "1.2.3 "
+badSemVs = [ -- Not enough version slots
+           "1", "1.2", "a.b.c"
+           -- Illegal characters
+           , "1.01.1", "1.2.3+a1b!2c3.1", "", "1.2.3 "
+           -- Really large version
+           -- , "18446744073709551610000000000000000.0.0"
            ]
 
 goodSemVs :: [T.Text]
@@ -54,6 +62,8 @@ goodSemVs = [ "0.1.0", "1.2.3", "1.2.3-1", "1.2.3-alpha", "1.2.3-alpha.2"
             , "1.0.0-alpha+001", "1.0.0+21AF26D3---117B344092BD"
             -- Zeroes
             , "1.2.2-00a"
+            -- Really large version
+            , "18446744073709551610.0.0"
             ]
 
 -- | The exact example from `http://semver.org`
@@ -78,8 +88,7 @@ suite :: TestTree
 suite = testGroup "Tests"
   [ testGroup "Unit Tests"
     [ testGroup "(Ideal) Semantic Versioning"
-      [ testGroup "Bad Versions (shouldn't parse)" $
-        map (\s -> testCase (T.unpack s) $ assertBool "A bad version parsed" $ isLeft $ semver s) badSemVs
+      [ testGroup "Bad Versions (shouldn't parse)" $ map bad badSemVs
       , testGroup "Good Versions (should parse)" $
         map (\s -> testCase (T.unpack s) $ isomorphSV s) goodSemVs
       , testGroup "Comparisons" $
@@ -97,6 +106,8 @@ suite = testGroup "Tests"
     , testGroup "(Haskell) PVP"
       [ testGroup "Good PVPs" $
         map (\s -> testCase (T.unpack s) $ isomorphPVP s) cabalOrd
+      , testGroup "Bad PVP" $
+        map (\s -> testCase (T.unpack s) $ assertBool "A bad PVP parsed" $ isLeft $ pvp s) badPVP
       , testGroup "Comparisons" $
         zipWith (\a b -> testCase (T.unpack $ a <> " < " <> b) $ comp pvp a b) cabalOrd (tail cabalOrd)
       ]
@@ -178,6 +189,10 @@ suite = testGroup "Tests"
         , eqVer "1.0.r15.g3fc772c-5"
         , eqVer "0.88-2"
         ]
+      , testGroup "Conversions"
+        [ testCase "Good Version -> PVP" $ versionToPvp $(versionQ "1.2.3") @?= Just $(pvpQ "1.2.3")
+        , testCase "Bad  Version -> PVP" $ versionToPvp $(versionQ "1.e.3") @?= Nothing
+        ]
       ]
     , testGroup "Lenses and Traversals"
       [ testCase "SemVer - Increment Patch" incPatch
@@ -185,18 +200,24 @@ suite = testGroup "Tests"
       , testCase "SemVer - Get patches" patches
       ]
     , testGroup "Template Haskell"
-      [ testCase "SemVer"  $ prettyV $(thVer "1.2.3") @?= "1.2.3"
-      , testCase "Version" $ prettyV $(thVer "1.2.3.4") @?= "1.2.3.4"
-      , testCase "Mess"    $ prettyV $(thVer "003.03-3") @?= "003.03-3"
-      , testCase "Failure" $ $(recover [| () |] (thVer "!!!")) @?= ()
+      [ testCase "SemVer"  $ prettyV $(versioningQ "1.2.3") @?= "1.2.3"
+      , testCase "Version" $ prettyV $(versioningQ "1.2.3.4") @?= "1.2.3.4"
+      , testCase "Mess"    $ prettyV $(versioningQ "003.03-3") @?= "003.03-3"
+      , testCase "Failure" $ $(recover [| () |] (versioningQ "!!!")) @?= ()
       ]
     , testGroup "Megaparsec Behaviour"
       [ testCase "manyTill" $ parse nameGrab "manyTill" "linux-firmware-3.2.14-1-x86_64.pkg.tar.xz" @?= Right "linux-firmware"
       , testCase "Extracting version" $ parse versionGrab "extraction" "linux-firmware-3.2.14-1-x86_64.pkg.tar.xz" @?= Right (Ideal $ SemVer 3 2 14 (Just . Release $ Alphanum "1-x86" :| []) Nothing)
+      , testCase "Parser State" $ parse pvp'' "parser state" "1.2.3arst" @?= Right ($(pvpQ "1.2.3"), "arst")
       ]
     ]
   ]
 
+bad :: T.Text -> TestTree
+bad s = testCase (T.unpack s) $ case semver s of
+  Left _   -> pure ()
+  Right s' -> assertFailure $ "A bad version parsed: " ++ T.unpack (prettySemVer s')
+
 compVer :: T.Text -> T.Text -> TestTree
 compVer a b = testCase (printf "%s < %s" a b) $ comp versioning a b
 
@@ -279,3 +300,10 @@ versionGrab = manyTill anySingle (try finished) *> ver
 hush :: Either a b -> Maybe b
 hush (Left _)  = Nothing
 hush (Right b) = Just b
+
+-- | An attempt to squeeze out the remaining parser state.
+pvp'' :: Parsec Void T.Text (PVP, T.Text)
+pvp'' = do
+  v <- pvp'
+  s <- getParserState
+  pure (v, stateInput s)
